module ECL where
-- from ghc:
import Data.List
import Data.Numbers.Primes
import Graphics.Gnuplot.Simple
-- local files:
import REL
import MODEXP
import RAND
import EREL (minimize,convertMapping)
import KRIPKEVIS

data Agent = Ag Integer deriving (Eq,Ord)
alice, bob, carol, dave, eve, mallory :: Agent
alice   = Ag 0; bob     = Ag 1;
carol   = Ag 2; dave    = Ag 3;
eve     = Ag 4; mallory = Ag 5

instance Enum Agent where
  fromEnum = (\(Ag n) -> fromIntegral n)
  toEnum   = (\n -> Ag (fromIntegral n))

instance Show Agent where
  show (Ag 0) = "Alice";  show (Ag 1) = "Bob"
  show (Ag 2) = "Carol";  show (Ag 3) = "Dave"
  show (Ag 4) = "Eve";    show (Ag 5) = "Mallory"
  show (Ag n) = ('a': show n)

data Prp = P Integer | Q Integer | R Integer | S Integer deriving (Eq,Ord)
instance Show Prp where
  show (P 0) = "p";  show (P n) = "p "++(show n)
  show (Q 0) = "q";  show (Q n) = "q "++(show n)
  show (R 0) = "r";  show (R n) = "r "++(show n);
  show (S 0) = "s";  show (S n) = "s "++(show n);

prpIndex :: Prp -> Integer
prpIndex (P k) = k*4
prpIndex (Q k) = k*4 + 1
prpIndex (R k) = k*4 + 2
prpIndex (S k) = k*4 + 3

type State      = Integer
type Partition  = [[State]]
type Register   = (Integer,Integer,[Integer])
type Constraint = (Prp,Prp)

registersize :: Integer
registersize = 2^(8::Int)

fullregister :: Register
fullregister = (0,registersize-1,[])

without :: Register -> Integer -> Register
without (low,high,excl) n = (low,high,nub (n:excl))

type Valuation = ([Prp],[Agent],[(Prp,Register)],[Constraint],[Constraint])

data CryptoM = Mo [State] [(Agent,Partition)] [(State,Valuation)] State deriving (Eq)

instance Show CryptoM where
  show (Mo sts rel val cur) = "(Mo "
    ++ show sts ++ "\n    "
    ++ show rel ++ "\n    "
    ++ show val ++ "\n    "
    ++ show cur ++ " )"

cm0for :: [Agent] -> CryptoM
cm0for ags = (Mo
  [0]
  [(a,[[0]]) | a <- ags]
  [ (0,([],[],[],[],[] )) ]
  0 )

states :: CryptoM -> [State]
states (Mo sts _ _ _) = sts
agents :: CryptoM -> [Agent]
agents (Mo _ rel _ _) = map fst rel
reachable :: CryptoM -> [State]
reachable model = nub $ concat $ map (reachableBy model) (agents model)
reachableBy :: CryptoM -> Agent -> [State]
reachableBy (Mo _ rel _ cur) agent
  = head $ filter (\set -> elem cur set) (apply rel agent)
reachableByFrom :: CryptoM -> Agent -> State -> [State]
reachableByFrom (Mo _ rel _ _) agent state
  = head $ filter (\set -> elem state set) (apply rel agent)
reachableFrom :: CryptoM -> State -> [State]
reachableFrom model state
  = nub $ concat $ map (\a -> reachableByFrom model a state) (agents model)
size :: CryptoM -> Int
size (Mo sts _ _ _) = length sts
facts :: CryptoM -> [Prp]
facts (Mo _ _ val cur) = fst5 (apply val cur)
factsAt :: CryptoM -> State -> [Prp]
factsAt (Mo _ _ val _) state = fst5 (apply val state)
listeners :: CryptoM -> [Agent]
listeners (Mo _ _ val cur) = snd5 (apply val cur)
listenersAt :: CryptoM -> State -> [Agent]
listenersAt (Mo _ _ val _) state = snd5 (apply val state)
nonlisteners :: CryptoM -> [Agent]
nonlisteners model = foldr delete (agents model) (listeners model)
nonlistenersAt :: CryptoM -> State -> [Agent]
nonlistenersAt model state = foldr delete (agents model) (listenersAt model state)
registers :: CryptoM -> [(Prp,Register)]
registers (Mo _ _ val cur) = trd5 (apply val cur)
registersAt :: CryptoM -> State -> [(Prp,Register)]
registersAt (Mo _ _ val _) state = trd5 (apply val state)
posConstraints :: CryptoM -> [Constraint]
posConstraints (Mo _ _ val cur) = fth5 (apply val cur)
posConstraintsAt :: CryptoM -> State -> [Constraint]
posConstraintsAt (Mo _ _ val _) state = fth5 (apply val state)
negConstraints :: CryptoM -> [Constraint]
negConstraints (Mo _ _ val cur) = fft5 (apply val cur)
negConstraintsAt :: CryptoM -> State -> [Constraint]
negConstraintsAt (Mo _ _ val _) state = fft5 (apply val state)

makeActual :: CryptoM -> State -> CryptoM
makeActual (Mo sts rel val _) newcur = if (elem newcur sts)
  then (Mo sts rel val newcur)
  else error ("World "++(show newcur)++" does not exist in this model!")

bisiMin :: CryptoM -> CryptoM
bisiMin (Mo oldstates oldrel oldval oldcur) = (Mo newstates newrel newval newcur)
  where
    newval         = nub $ map (\(x,v) -> (apply bisim x, v)) newvalEntries
    newvalEntries  = filter (\x -> (elem (apply bisim (fst x)) newstates)) oldval
    newstates      = nub $ map (apply bisim) oldstates
    (newrel,bisim) = convertMapping [0..] $ minimize oldrel oldval
    newcur         = apply bisim oldcur

genMin :: CryptoM -> CryptoM
genMin model@(Mo _ oldrel oldval cur) = (Mo newstates newrel newval cur)
  where
    newstates   = lfp (\set -> mark set) (reachable model)
    newval      = filter (\x -> elem (fst x) newstates) oldval
    newrel      = [ (a, newrelfor a) | a <- agents model ]
    newrelfor a = filter (\part -> elem (head part) newstates) (apply oldrel a)
    mark marked = nub $ concat $ map (reachableFrom model) marked

data Form = Top | PrpF Prp | L Agent | Equal Exp Exp
  | Neg Form | Conj [Form]
  | K Agent Form | G Form
  | Com Com Form
  | Prime Exp | Coprime Exp Exp
  deriving (Eq,Ord,Show)

data Com =  Open Agent | Close Agent
  | Create Prp Agent Exp | CreateSized Prp Agent Exp Integer
  | Announce Prp | AnnounceEqual Prp Exp | AnnounceNotEqual Prp Exp
  | Test Form | Com :- Com
  deriving (Eq,Ord,Show)

data Exp = PrpE Prp | Nmbr Integer
  | PlusMod Exp Exp Exp | TimesMod Exp Exp Exp | PowerMod Exp Exp Exp
  deriving (Eq,Ord,Show)

bot :: Form
bot = Neg Top

disj :: [Form] -> Form
disj list = Neg $ Conj [ Neg d | d <- list ]

implies :: Form -> Form -> Form
implies a b = disj [Neg a, b]

box :: Com -> Form -> Form
box com form = Neg ( Com com ( Neg form ) )

lst2cmd :: [Com] -> Com
lst2cmd [] = error "empty list"
lst2cmd [c] = c
lst2cmd [c1,c2] = c1 :- c2
lst2cmd (c1:c2s) = c1 :- (lst2cmd c2s)

propsInForm :: Form -> [Prp]
propsInForm Top             = []
propsInForm (PrpF aprop)    = [aprop]
propsInForm (Neg formula)   = propsInForm formula
propsInForm (Conj forms)    = nub $ concat (map propsInForm forms)
propsInForm (K _ formula)   = propsInForm formula
propsInForm (G formula)     = propsInForm formula
propsInForm (L _ )          = []
propsInForm (Com c formula) = nub $ (propsInForm formula) ++ (propsInCom c)
propsInForm (Equal a b)     = nub $ propsInExp a ++ propsInExp b
propsInForm (Prime a)       = propsInExp a
propsInForm (Coprime a b)   = nub $ propsInExp a ++ propsInExp b

propsInExp :: Exp -> [Prp]
propsInExp (PrpE aprop)     = [aprop]
propsInExp (Nmbr _)         = []
propsInExp (PlusMod  a b c) = nub $ concat $ map propsInExp [a,b,c]
propsInExp (TimesMod a b c) = nub $ concat $ map propsInExp [a,b,c]
propsInExp (PowerMod a b c) = nub $ concat $ map propsInExp [a,b,c]

propsInCom :: Com -> [Prp]
propsInCom (Open _)               = []
propsInCom (Close _)              = []
propsInCom (Create p _ e)         = nub $ [p] ++ propsInExp e
propsInCom (CreateSized p _ e _)  = nub $ [p] ++ propsInExp e
propsInCom (Announce p)           = [p]
propsInCom (AnnounceEqual p e)    = nub $ [p] ++ propsInExp e
propsInCom (AnnounceNotEqual p e) = nub $ [p] ++ propsInExp e
propsInCom (comA :- comB)         = nub $ propsInCom comA ++ propsInCom comB
propsInCom (Test f)               = nub $ propsInForm f

type Assignment = [(Prp,Integer)]
evalEAss :: Assignment -> Exp -> Integer
evalEAss _   (Nmbr n) = n
evalEAss ass (PrpE p) = apply ass p
evalEAss ass (PlusMod a b m)  =
  mod ( (evalEAss ass a) + (evalEAss ass b) ) (evalEAss ass m)
evalEAss ass (TimesMod a b m) =
  mod ( (evalEAss ass a) * (evalEAss ass b) ) (evalEAss ass m)
evalEAss ass (PowerMod a b m) =
  exM (evalEAss ass a) (evalEAss ass b) (evalEAss ass m)

consistent :: [Constraint] -> [Constraint] -> Assignment -> Bool
consistent pcs ncs ass = and [all equal pcs, all (not.equal) ncs]
  where
    equal (p1,p2) = ( (apply ass p1) == (apply ass p2) )

allAss :: CryptoM -> [Assignment]
allAss model = filter (consistent pcs ncs) (aALoop [] (registers model))
  where
    pcs = posConstraints model
    ncs = negConstraints model

aALoop :: [ Assignment ] -> [ (Prp,Register) ] -> [ Assignment ]
aALoop []   []     = [ [] ]
aALoop done []     = done
aALoop []   (x:xs) = aALoop [ [ ((fst x),v) ] | v <- reg2lst (snd x) ] xs
aALoop done (x:xs) = aALoop [ (((fst x),v):o) | v <- reg2lst (snd x), o <- done ] xs

reg2lst :: Register -> [Integer]
reg2lst (low,high,excl) = foldr delete [low..high] excl

allRelevantAss :: CryptoM -> [Prp] -> [Assignment]
allRelevantAss model props =
  filter (consistent pcs ncs) (aALoop [] (restrict (registers model) relprops))
    where
      relprops = nub $ props ++ (\l -> (map fst l)++(map snd l)) (pcs++ncs)
      pcs      = posConstraints model
      ncs      = negConstraints model

restrict :: Eq a => [(a,b)] -> [a] -> [(a,b)]
restrict rel domain = filter (\pair -> elem (fst pair) domain) rel

evalAss :: CryptoM -> Assignment -> Form -> Bool

evalAss _     _   Top           = True
evalAss model _   (PrpF prp)    = elem prp (facts model)
evalAss _     ass (Equal a b)   = (evalEAss ass a) == (evalEAss ass b)
evalAss _     ass (Prime e)     = isPrime (evalEAss ass e)
evalAss _     ass (Coprime a b) = (gcd (evalEAss ass a) (evalEAss ass b) == 1)
evalAss model ass (Neg form)    = not (evalAssMC model ass form)
evalAss model ass (Conj forms)  = and (map (evalAssMC model ass) forms)

evalAss model _ (K agent form) = and results
  where results   = map evalthere (reachableBy model agent)
	evalthere = (\v -> (evalAt model v form==Just True))

evalAss model _ (G form) = and results
  where results   = map evalthere (states (genMin model))
	evalthere = (\v -> (evalAt model v form==Just True))

evalAss model _ (L agent) = elem agent (listeners model)

evalAss model ass (Com com form) =
  if (assSet /= [])
    then and results
    else error ("No compatible assignments!")
  where
    newmodel = update model com
    assSet   = filter (subs ass) (allRelevantAss newmodel props)
    props    = nub $ propsInCom com ++ propsInForm form
    chkFct   = (\newass -> evalAss (newmodel) newass form)
    results  = map chkFct assSet
    subs a b = all (\x -> (apply a x == apply b x)) (map fst a)

eval :: CryptoM -> Form -> Maybe Bool
eval model formula =
  if (and results)
    then
      Just True
    else
      if (and $ map not results)
	then Just False
	else Nothing
  where
    results = [ evalAss model ass formula | ass <- assSet ]
    assSet  = allRelevantAss model (propsInForm formula)

evalAt :: CryptoM -> State -> Form -> Maybe Bool
evalAt (Mo sts rel val _) newcur form = eval (Mo sts rel val newcur) form

evalAtAll :: CryptoM -> Form -> [(State,Maybe Bool)]
evalAtAll m form = zip (states m) (map (\w -> evalAt m w form) (states m))

type ValChange = Valuation -> Valuation
type ActionS = ( [State], [(State,Form)], [(State,ValChange)], [(Agent,Partition)] )
type Action = (ActionS,State)

productUpdateWithEvAtFct ::
  (CryptoM -> State -> Form -> Maybe Bool) -> CryptoM -> Action -> CryptoM
productUpdateWithEvAtFct evAtFct model (actionStructure,faction) =
  let
    (Mo oldstates oldrel oldval oldcur) = model
    (actions, tests, changes, actrel)   = actionStructure
    startcount        = (maximum oldstates) + 1
    copiesOf (s,a)    = if (evAtFct model s (apply tests a) == Just True)
			  then [ (s,a,(a*startcount + s)) ]
			  else [ ]
    newstatesTriples  = concat [ copiesOf (s,a) | s <- oldstates, a <- actions ]
    newstates         = map trd3 newstatesTriples
    newValFor (s,a,t) = (t, (apply changes a) (apply oldval s))
    newval            = map newValFor newstatesTriples
    listFor ag        = cartProd (apply oldrel ag) (apply actrel ag)
    newPartsFor ag    = [ cartProd as bs | (as,bs) <- listFor ag ]
    translSingle pair = filter (\x-> elem x newstates) $ map trd3 $ copiesOf (pair)
    transEqClass list = concat $ map translSingle list
    nTransPartsFor ag = filter (\x-> x/=[]) $ map transEqClass (newPartsFor ag)
    newrel            = [ (a, nTransPartsFor a)  | a <- (agents model) ]
    newcur            = trd3 $ head $ copiesOf (oldcur,faction)
    factTest          = apply tests faction
  in
    if (sort $ nub (agents model)) == (sort $ nub (map fst actrel))
      then if (evAtFct model oldcur factTest == Just True)
	then genMin $ bisiMin $ (Mo newstates newrel newval newcur)
	else error ("Actual precondition '" ++ (show factTest) ++ "' is false!")
      else error "Agents of model and actionStructure are not the same!"

productUpdate :: CryptoM -> Action -> CryptoM
productUpdate = productUpdateWithEvAtFct evalAt

openAction, closeAction :: CryptoM -> Agent -> Action
openAction model a = (([0,1], [(0,Top),(1,Top)], [(0,id),(1,addLst)], actrel), 1)
  where
    addLst = \(fs,lstnrs,reg,pcs,ncs) -> (fs,nub(a:lstnrs),reg,pcs,ncs)
    actrel = [ (i,[[0],[1]]) | i <- hear  ] ++ [ (i,[[0,1]]) | i <- dumb ]
    hear   = nub (a : listeners model)
    dumb   = foldr delete (agents model) hear

closeAction model a = (([0,1], [(0,Top),(1,Top)], [(0,id),(1,remLst)], actrel), 1)
  where
    remLst = \(fs,lstnrs,reg,pcs,ncs) -> (fs,delete a lstnrs,reg,pcs,ncs)
    hear   = nub (a : listeners model)
    dumb   = foldr delete (agents model) hear
    actrel = [ (i,[[0],[1]]) | i <- hear  ] ++ [ (i,[[0,1]]) | i <- dumb ]

announceAction :: CryptoM -> Prp -> Action
announceAction model p = ( ( [0,1],
  [ (0,PrpF p), (1,Neg $ (PrpF p)) ],
  [ (0,id    ), (1,id            ) ], actrel ), 0 )
  where
    actrel = [ (j,[[0],[1]]) | j <- (listeners model)    ]
	  ++ [ (j,[ [0,1] ]) | j <- (nonlisteners model) ]

announceEqualNAction, announceNotEqualNAction :: CryptoM -> Prp -> Integer -> Action
announceEqualNAction model p n = ( ( [0,1],
  [ (0,Equal (PrpE p) (Nmbr n)), (1,Neg $ Equal (PrpE p) (Nmbr n)) ],
  [ (0,id                     ), (1,id                           ) ], actrel ), 0 )
  where
    actrel = [ (j,[[0],[1]]) | j <- (listeners model)    ]
	  ++ [ (j,[ [0,1] ]) | j <- (nonlisteners model) ]

announceEqualPAction, announceNotEqualPAction :: CryptoM -> Prp -> Prp -> Action
announceEqualPAction model p q = ( ( [0,1,2],
  [ (0,Conj[PrpF p,PrpF q,Equal (PrpE p) (PrpE q)]),
    (1,Conj[Neg (PrpF p),Neg (PrpF q)]),
    (2,Top) ],
  [ (0,id), (1,addPC), (2,id) ], actrel ), 0 )
  where
    addPC  = \(fcts,ls,regs,pc,nc) -> (fcts,ls,regs,(p,q):pc,nc)
    actrel = [ (j,[[0,1],[2]]) | j <- (listeners model)    ]
	  ++ [ (j,[ [0,1,2] ]) | j <- (nonlisteners model) ]

announceNotEqualNAction model p n = ( ( [0,1],
  [(0,Conj [PrpF p, Neg $ Equal (PrpE p) (Nmbr n)]),
   (1,Neg (PrpF p)),
   (2,Top) ],
  [(0,id), (1,exclN), (2,id) ], actrel ), 0 )
  where
    exclN  = \(fcts,ls,regs,ncs,pcs) -> (fcts,ls, map change regs,ncs,pcs)
    change (prp,reg) = if (prp == p) then (prp,without reg n) else (prp,reg)
    actrel = [ (j,[[0,1],[2]]) | j <- (listeners model)    ]
	  ++ [ (j,[ [0,1,2] ]) | j <- (nonlisteners model) ]

announceNotEqualPAction model p q = ( ( [0,1,2],
  [(0,Conj[PrpF p, PrpF q, Neg $ Equal (PrpE p) (PrpE q)]),
   (1,Neg(Conj[PrpF p,PrpF q])),
   (2,Top)],
  [(0,id), (1,addNC), (2,id)], actrel ), 0 )
  where
    addNC  = \(fcts,ls,regs,pcs,ncs) -> (fcts,ls,regs,pcs,((p,q):ncs))
    actrel = [ (j,[[0,1],[2]]) | j <- (listeners model)    ]
	  ++ [ (j,[ [0,1,2] ]) | j <- (nonlisteners model) ]

testAction :: CryptoM -> Form -> Action
testAction model form = ( ( [0,1],
  [ (0,form), (1,Neg form) ],
  [ (0,id  ), (1,id      ) ], actrel ), 0 )
  where
    actrel = [ (j,[[0,1]]) | j <- (agents model) ]

createSizedAction :: CryptoM -> Prp -> Agent -> Exp -> Integer -> Action
createSizedAction model p i e regmax = ( ( [0,1],
  [ (0, pre   ), (1, pre   ) ],
  [ (0, addFct), (1, addReg) ],
  actrel ), 0 )
  where
    assSet = allRelevantAss model (propsInExp e)
    ass    = if (assSet /= []) then (head assSet) else error "No assignment."
    n      = evalEAss ass e
    pre    = G (Neg (PrpF p))
    addFct = \(fcts,ls,reg,pc,nc) -> (p:fcts,ls,(p,(n,n     ,[ ])):reg,pc,nc)
    addReg = \(fcts,ls,reg,pc,nc) -> (  fcts,ls,(p,(0,regmax,[n])):reg,pc,nc)
    others = delete i (agents model)
    actrel = [ (i,[[0],[1]]) ] ++ [ (j,[[0,1]]) | j <- others ]

update :: CryptoM -> Com -> CryptoM
update model (com1 :- com2) = update (update model com1) com2
update model (Open a)       = productUpdate model (openAction model a)
update model (Close a)      = productUpdate model (closeAction model a)
update model (Announce p)   = productUpdate model (announceAction model p)
update model (Test form)    = productUpdate model (testAction model form)
update model (Create p i e) = update model (CreateSized p i e registersize)
update model (CreateSized p i e regmax) =
  productUpdate model (createSizedAction model p i e regmax)
update model (AnnounceEqual p (Nmbr n)) =
  productUpdate model (announceEqualNAction model p n)
update model (AnnounceEqual p (PrpE q)) =
  productUpdate model (announceEqualPAction model p q)
update model (AnnounceNotEqual p (Nmbr n)) =
  productUpdate model (announceNotEqualNAction model p n)
update model (AnnounceNotEqual p (PrpE q)) =
  productUpdate model (announceNotEqualPAction model p q)
update _ _ = error "Update is not defined for this expression."

stepwiseUpdateEval :: CryptoM -> [Com] -> Form -> String
stepwiseUpdateEval model []     f = (show f) ++ " is " ++ (show $ eval model f)
stepwiseUpdateEval model (x:xs) f = "After "++(show x)++":"
  ++"\n size="++(show $ size nextm)
  ++"\n regs="++(show $ map (\(p,reg) -> (p,(head $ reg2lst reg))) (registers nextm))
  ++"\n\n"++(stepwiseUpdateEval nextm xs f)
  where nextm = update model x

assAmount :: Integer
assAmount = 2

myRandSeed :: [Integer]
myRandSeed = concat $ [ map (n+) myRandSeed10000 | n <- [0..] ]

rndRelAssSingle :: CryptoM -> [Prp] -> Integer -> Assignment
rndRelAssSingle model givenprops seed = [ (p, pickFor p) | p <- props ]
  where
    modprops  = map fst (registers model)
    props     = filter (\x -> (elem x modprops)) givenprops
    domFor p  = reg2lst (apply (registers model) p)
    lenFor p  = fromIntegral $ length (domFor p)
    seedFor p = mod (seed*(myRandSeed!!(fromIntegral $ (prpIndex p+seed)))) (lenFor p)
    pickFor p = (domFor p)!!(fromIntegral $ seedFor p)

rndRelAss :: CryptoM -> [Prp] -> Integer -> [Assignment]
rndRelAss model props amount = take (fromIntegral amount) consSet
  where
    pcs = posConstraints model
    ncs = negConstraints model
    consSet = filter (consistent pcs ncs) fullSet
    fullSet = [ rndRelAssSingle model props seed | seed <- [1..] ]

rndTestModel :: Integer -> CryptoM
rndTestModel n = ( Mo [0] [ (alice,[[0]]) ]
  [ (0,([],[],[ ((P 0),(0,n,[0])), ((Q 0),(0,n,[0]))  ],[],[])) ] 0 )

rndTestRun :: Integer -> Integer -> IO ()
rndTestRun rs amount = plotListStyle arg1 arg2 coords
  where
    arg1 = [EPS filename, XRange (0,fromIntegral rs), YRange (0,fromIntegral rs)]
    arg2 = (defaultStyle {plotType = Points, lineSpec = CustomStyle [LineTitle ((show amount)++" assignments for two variables")]})
    coords   = map (\list -> (head list, head $ tail list)) numbers
    numbers  = map (\list -> (map snd list)) assSet
    assSet   = rndRelAss model [(P 0),(Q 0)] amount
    model    = (rndTestModel rs)
    filename = "img/rndtest_"++(show rs)++"_regsize_"++(show amount)++"ass.eps"

rndTestAllRuns :: IO ()
rndTestAllRuns = do
  rndTestRun (2^(8::Int)) 100
  rndTestRun (2^(8::Int)) 1000
  rndTestRun (2^(8::Int)) 3000

evalAssMC :: CryptoM -> Assignment -> Form -> Bool

evalAssMC model _   (PrpF prp)    = elem prp (facts model)
evalAssMC _     ass (Equal a b)   = (evalEAss ass a) == (evalEAss ass b)
evalAssMC _     ass (Prime e)     = isPrime (evalEAss ass e)
evalAssMC _     ass (Coprime a b) = (gcd (evalEAss ass a) (evalEAss ass b) == 1)
evalAssMC _     _   Top           = True
evalAssMC model ass (Neg form)    = not (evalAssMC model ass form)
evalAssMC model ass (Conj forms)  = and (map (evalAssMC model ass) forms)

evalAssMC model _   (K i form)    = and results
  where results = map (\v -> (evalMCAt model v form==Just True)) (reachableBy model i)

evalAssMC model _   (G form)      = and results
  where results = map (\v -> (evalMCAt model v form==Just True)) (states $ genMin model)

evalAssMC model _   (L agent)     = elem agent (listeners model)

evalAssMC model ass (Com com form) =
  if (assSet /= [])
    then and results
    else error ("No compatible assignments!")
  where
    newmodel = updateMC model com
    assSet   = filter (subset ass) (rndRelAss newmodel props assAmount)
    props    = nub $ propsInCom com ++ propsInForm form
    chkFct   = (\newass -> evalAssMC newmodel newass form)
    results  = map chkFct assSet
    subset assA assB = all (\x -> (apply assA x == apply assB x)) (map fst assA)

evalMC :: CryptoM -> Form -> Maybe Bool
evalMC model formula =
    if (and results)
      then
	Just True
      else
	if (all (\x -> x==False) results)
	  then Just False
	  else Nothing
  where
    results = [ evalAssMC model ass formula | ass <- assSet ]
    assSet  = rndRelAss model (propsInForm formula) assAmount

evalMCAt :: CryptoM -> State -> Form -> Maybe Bool
evalMCAt (Mo sts rel val _) nc f = evalMC (Mo sts rel val nc) f

evalMCAtAll :: CryptoM -> Form -> [(State,Maybe Bool)]
evalMCAtAll m form = zip (states m) (map (\w -> evalMCAt m w form) (states m))

productUpdateMC :: CryptoM -> Action -> CryptoM
productUpdateMC = productUpdateWithEvAtFct evalMCAt

updateMC :: CryptoM -> Com -> CryptoM
updateMC model (com1 :- com2) = updateMC (updateMC model com1) com2
updateMC model (Open a)       = productUpdateMC model (openAction model a)
updateMC model (Close a)      = productUpdateMC model (closeAction model a)
updateMC model (Announce p)   = productUpdateMC model (announceAction model p)
updateMC model (Test form)    = productUpdateMC model (testAction model form)
updateMC model (Create p i e) = updateMC model (CreateSized p i e registersize)
updateMC model (CreateSized p i e regmax) =
  productUpdateMC model (createSizedAction model p i e regmax)
updateMC model (AnnounceEqual p (Nmbr n)) =
  productUpdateMC model (announceEqualNAction model p n)
updateMC model (AnnounceEqual p (PrpE q)) =
  productUpdateMC model (announceEqualPAction model p q)
updateMC model (AnnounceNotEqual p (Nmbr n)) =
  productUpdateMC model (announceNotEqualNAction model p n)
updateMC model (AnnounceNotEqual p (PrpE q)) =
  productUpdateMC model (announceNotEqualPAction model p q)
updateMC _ _ = error "Update is not defined for this expression."

stepwiseUpdateEvalMC :: CryptoM -> [Com] -> Form -> String
stepwiseUpdateEvalMC model []     f = (show f) ++ " is " ++ (show $ evalMC model f)
stepwiseUpdateEvalMC model (x:xs) f = "After "++(show x)++":"
  ++"\n  size = "++(show $ size nextmodel)
  ++" // regs = "++(show $ map (\(p,reg) -> (p,(head $ reg2lst reg))) (registers nextmodel))
  ++"\n\n"++(stepwiseUpdateEvalMC nextmodel xs f)
  where nextmodel = updateMC model x

eclShowProp :: Prp -> String
eclShowProp prp = replace (replace (show prp) " 0" "") " " "_"

eclShowConstraints :: [Constraint] -> [Constraint] -> String
eclShowConstraints []  []  = ""
eclShowConstraints pcs ncs =
  sepBy (positives ++ negatives) " \\text{ and } "
    where
      positives = map (niceconSingle "   =  ") pcs
      negatives = map (niceconSingle " \\neq ") ncs
      niceconSingle b (prpA,prpB) = " $ " ++ (eclShowProp prpA) ++ b ++ (eclShowProp prpB) ++ " $ "

eclShowVal :: Valuation -> String
eclShowVal (fcts,lstnrs,regs,pcs,ncs) =
  sepBy nonEmptyItems newline
    where
      nonEmptyItems = filter (\x -> x/="") items
      items = [nicefacts, nicelisteners, nicereg, eclShowConstraints pcs ncs]
      nicelisteners = sepBy (map show lstnrs) ", "
      nicefacts     = sepBy (map show fcts) ", "
      nicereg       = concat $ map eclShowReg regs

eclShowReg :: (Prp,Register) -> String
eclShowReg (prp,(low,high,excl)) =
  if (low /= high)
    then
      " $ " ++ (show low) ++ " \\leq " ++ (show prp) ++ " \\leq " ++ (show high) ++ "\\text{ and }"++(show prp)++"\\not\\in \\{" ++ (concat $ map show excl) ++ "\\} $ " ++ newline
    else
      " $ " ++ (show prp) ++ "=" ++ (show low) ++  " $ " ++ newline

eclInfo :: CryptoM -> String
eclInfo model = begintab ++ "Agents: " ++ (sepBy (map show (agents model)) ", ") ++ endtab

eclTexModel :: CryptoM -> String -> IO String
eclTexModel model = texModel show show eclShowVal (eclInfo model) visModel
  where
    (Mo sts rel val cur) = model
    visModel = (VisModel sts rel val cur)

eclDispModel :: CryptoM -> IO String
eclDispModel model = dispModel show show eclShowVal (eclInfo model) visModel
  where
    (Mo sts rel val cur) = model
    visModel = (VisModel sts rel val cur)

