module GG where
-- from ghc:
import Data.List
-- local files:
import REL
import KRIPKEVIS

data Agent = Ag Integer deriving (Eq,Ord)
jan,gaia,rosa :: Agent
jan  = Ag 0
gaia = Ag 1
rosa = Ag 2

instance Enum Agent where
  fromEnum = (\(Ag n) -> fromIntegral n)
  toEnum = (\n -> Ag (fromIntegral n))

instance Show Agent where
  show (Ag 0) = "Jan"
  show (Ag 1) = "Gaia"
  show (Ag 2) = "Rosa"
  show (Ag n) = "Ag "++(show n)

data Prp = P Integer deriving (Eq,Ord)

instance Show Prp where
  show (P 0) = "p";
  show (P n) = "p "++(show n);

prpIndex :: Prp -> Integer
prpIndex (P k) = k

type State = Integer

type Partition =  [[State]]

type Register = (Integer,Integer,[Integer])

fullregister :: Register
fullregister = (1,10,[])

without :: Register -> Integer -> Register
without (low,high,excl) n = (low,high,nub (n:excl))

type Constraint = (Prp,Prp)

type Valuation = ([Prp],[(Prp,Register)],[Constraint],[Constraint])

data GuessM = Mo
  [State]
  [(Agent,Partition)]
  [(State,Valuation)]
  State
  deriving (Eq)

instance Show GuessM where                                                                                       
  show (Mo sts rel val cur) = "(Mo \n  "
    ++ show sts ++ "\n  "
    ++ show rel ++ "\n  "
    ++ show val ++ "\n  "
    ++ show cur ++ "\n  )"

m0for :: [Agent] -> GuessM
m0for ags = (Mo
  [0]
  [(a,[[0]]) | a <- ags]
  [ (0,([],[],[],[] )) ]
  0
  )

agents :: GuessM -> [Agent]
agents (Mo _ rel _ _) = map fst rel
states :: GuessM -> [State]
states (Mo s _ _ _) = s
reachable :: GuessM -> [State]
reachable model = nub $ concat $ map (reachableBy model) (agents model)
reachableBy :: GuessM -> Agent -> [State]
reachableBy (Mo _ rel _ cur) agent
  = head $ filter (\set -> elem cur set) (apply rel agent)
reachableByFrom :: GuessM -> Agent -> State -> [State]
reachableByFrom (Mo _ rel _ _) agent state
  = head $ filter (\set -> elem state set) (apply rel agent)
reachableFrom :: GuessM -> State -> [State]
reachableFrom model state
  = nub $ concat $ map (\a -> reachableByFrom model a state) (agents model)
size :: GuessM -> Int
size (Mo sts _ _ _) = length sts
facts :: GuessM -> [Prp]
facts (Mo _ _ val cur) = fst4 (apply val cur)
factsAt :: GuessM -> State -> [Prp]
factsAt (Mo _ _ val _) state = fst4 (apply val state)
registers :: GuessM -> [(Prp,Register)]
registers (Mo _ _ val cur) = snd4 (apply val cur)
registersAt :: GuessM -> State -> [(Prp,Register)]
registersAt (Mo _ _ val _) state = snd4 (apply val state)
posConstraints :: GuessM -> [Constraint]
posConstraints (Mo _ _ val cur) = trd4 (apply val cur)
posConstraintsAt :: GuessM -> State -> [Constraint]
posConstraintsAt (Mo _ _ val _) state = trd4 (apply val state)
negConstraints :: GuessM -> [Constraint]
negConstraints (Mo _ _ val cur) = fth4 (apply val cur)
negConstraintsAt :: GuessM -> State -> [Constraint]
negConstraintsAt (Mo _ _ val _) state = fth4 (apply val state)

data Form = Top | PrpF Prp | Equal Prp Exp
  | Neg Form | Conj [Form]
  | K Agent Form | G Form | Com Com Form
  deriving (Eq,Ord,Show)

data Com = AnnounceEqual Prp Exp
  | AnnounceNotEqual Prp Exp
  | Create Prp Agent Integer
  | Com :- Com
  deriving (Eq,Ord,Show)

data Exp = PrpE Prp | Nmbr Integer
  deriving (Eq,Ord,Show)

bot :: Form
bot = Neg Top

disj :: [Form] -> Form
disj list = Neg $ Conj [ Neg d | d <- list ]

implies :: Form -> Form -> Form
implies a b = disj [Neg a, b]

box :: Com -> Form -> Form
box com form = Neg (Com com (Neg form))

propsInForm :: Form -> [Prp]
propsInForm Top             = []
propsInForm (PrpF aprop)    = [aprop]
propsInForm (Equal p e)     = nub $ [p] ++ propsInExp e
propsInForm (Neg formula)   = propsInForm formula
propsInForm (Conj formset)  = nub $ concat (map propsInForm formset)
propsInForm (K _ formula)   = propsInForm formula
propsInForm (G formula)     = propsInForm formula
propsInForm (Com c formula) = nub $ (propsInForm formula) ++ (propsInCom c)

propsInExp :: Exp -> [Prp]
propsInExp (PrpE aprop)     = [aprop]
propsInExp (Nmbr _)         = []

propsInCom :: Com -> [Prp]
propsInCom (Create p _ _)         = [p]
propsInCom (AnnounceEqual p e)    = nub $ [p] ++ propsInExp e
propsInCom (AnnounceNotEqual p e) = nub $ [p] ++ propsInExp e
propsInCom (com1 :- com2)         = nub $ propsInCom com1 ++ propsInCom com2

type Assignment = [(Prp,Integer)]

evalEAss :: Assignment -> Exp -> Integer
evalEAss _   (Nmbr n) = n
evalEAss ass (PrpE p) = apply ass p

consistent :: [Constraint] -> [Constraint] -> Assignment -> Bool
consistent pcs ncs ass = and [all equal pcs, all (not.equal) ncs]
  where
    equal (p1,p2) = ( (apply ass p1) == (apply ass p2) )

allAss :: GuessM -> [Assignment]
allAss model = filter (consistent pcs ncs) (aALoop [] (registers model))
  where
    pcs = posConstraints model
    ncs = negConstraints model

aALoop :: [ Assignment ] -> [ (Prp,Register) ] -> [ Assignment ]
aALoop []   []     = [ [] ]
aALoop done []     = done
aALoop []   (x:xs) = aALoop [ [ ((fst x),v) ] | v <- reg2lst (snd x) ] xs
aALoop done (x:xs) = aALoop [ (((fst x),v):o) | v <- reg2lst (snd x), o <- done ] xs

reg2lst :: Register -> [Integer]
reg2lst (low,high,excl) = foldr delete [low..high] excl

allRelevantAss :: GuessM -> [Prp] -> [Assignment]
allRelevantAss model props =
  filter (consistent pcs ncs) (aALoop [] (restrict (registers model) relprops))
    where
      relprops = nub $ props ++ (\l -> (map fst l)++(map snd l)) (pcs++ncs)
      pcs = posConstraints model
      ncs = negConstraints model

restrict :: Eq a => [(a,b)] -> [a] -> [(a,b)]
restrict rel domain = filter (\pair -> elem (fst pair) domain) rel

evalAss :: GuessM -> Assignment -> Form -> Bool

evalAss model _ (PrpF p)    = (elem p (facts model))

evalAss _ ass (Equal p e)   = (a == b)
  where a = evalEAss ass (PrpE p)
        b = evalEAss ass e

evalAss _     _   Top       = True
evalAss model ass (Neg f)   = not (evalAss model ass f)
evalAss model ass (Conj fs) = and (map (evalAss model ass) fs)

evalAss model _ (G f) = all (\x -> x == Just True) set
  where set = map (\s -> evalAt model s f) (states model)

evalAss model _ (K agent f) = all (\x -> x == Just True) set
  where set = map (\s -> evalAt model s f) (reachableBy model agent)

evalAss model ass (Com com form) =
  if (assSet /= [])
    then and results
    else error ("No compatible assignments!")
  where
    newmodel = update model com
    assSet   = filter (subs ass) (allRelevantAss newmodel props)
    props    = nub $ propsInCom com ++ propsInForm form
    chkFct   = (\newass -> evalAss (newmodel) newass form)
    results  = map chkFct assSet
    subs a b = all (\x -> (apply a x == apply b x)) (map fst a)

eval :: GuessM -> Form -> Maybe Bool
eval model formula =
  if (and results)
    then
      Just True
    else
      if (and $ map not results)
	then Just False
	else Nothing
  where
    results = [ evalAss model ass formula | ass <- assSet ]
    assSet  = allRelevantAss model (propsInForm formula)

evalAt :: GuessM -> State -> Form -> Maybe Bool
evalAt (Mo sts rel val _) newcur form = eval (Mo sts rel val newcur) form

type ValChange = Valuation -> Valuation

type ActionS   = ( [State], [(State,Form)], [(State,ValChange)], [(Agent,Partition)] )

type Action    = ( ActionS, State )

productUpdate :: GuessM -> Action -> GuessM
productUpdate model@(Mo oldstates oldrel oldval oldcur) (actionStructure,faction) =
  let
    (actions, tests, changes, actrel) = actionStructure
    startcount        = (maximum oldstates) + 1
    newstatesTriples  = concat [ copiesOf (s,a) | s <- oldstates, a <- actions ]
    copiesOf (s,a)    = if (evalAt model s (apply tests a) == Just True)
			  then [ (s,a,(a*startcount + s)) ]
			  else [ ]
    newstates         = map trd3 newstatesTriples
    newValFor (s,a,t) = (t, (apply changes a) (apply oldval s))
    newval            = map newValFor newstatesTriples
    listFor a         = cartProd (apply oldrel a) (apply actrel a)
    newPartsFor a     = [ cartProd as bs | (as,bs) <- listFor a ]
    translSingle pair = map trd3 $ take 1 (copiesOf (pair))
    transEqClass list = concat $ map translSingle list
    nTransPartsFor a  = map transEqClass (newPartsFor a)
    newrel            = [ (a, nTransPartsFor a)  | a <- (agents model) ]
    newcur            = trd3 $ head $ copiesOf (oldcur,faction)
    factTest          = apply tests faction
  in
    if (sort $ nub (agents model)) == (sort $ nub (map fst actrel))
      then if (eval model factTest == Just True)
        then (Mo newstates newrel newval newcur)
	else error ("The actual precondition '" ++ (show factTest) ++ "' is false!")
    else error "Agent sets of model and actionStructure are not the same!"

update :: GuessM -> Com -> GuessM

update model (AnnounceEqual p (Nmbr n)) = productUpdate model action
  where
    action = ( ( [0],
      [(0,Equal p (Nmbr n))],
      [(0,id              )],
      actrel ), 0 )
    actrel = [ (i,[[0]]) | i <- (agents model) ]

update model (AnnounceEqual p (PrpE q)) = productUpdate model action
  where
    action = ( ( [0,1],
      [(0,Conj[PrpF p,PrpF q,Equal p (PrpE q)]),(1,Conj[Neg (PrpF p),Neg (PrpF q)])],
      [(0,id                                  ),(1,addPC                          )],
      actrel ), 0 )
    addPC  = \(fcts,regs,pc,nc) -> (fcts,regs,(p,q):pc,nc)
    actrel = [ (i,[[0,1]]) | i <- (agents model) ]

update model (AnnounceNotEqual p (Nmbr n)) = productUpdate model action
  where
    action = ( ( [0,1],
      [(0,Conj [PrpF p, Neg $ Equal p (Nmbr n)]), (1,Neg (PrpF p))],
      [(0,id                                   ), (1,exclN       )],
      actrel ), 0 )
    exclN  = \(fcts, regs, nc, pc) -> (fcts, map change regs, nc, pc)
    change (prp,reg) = if (prp == p) then (prp,without reg n) else (prp,reg)
    actrel = [ (i,[[0,1]]) | i <- (agents model) ]

update model (AnnounceNotEqual p (PrpE q)) = productUpdate model action
  where
    action = ( ( [0,1],
      [(0,Conj[PrpF p,PrpF q,Neg$Equal p (PrpE q)]),(1,Neg(Conj[PrpF p,PrpF q]))],
      [(0,id                                      ),(1,addNC                   )],
      actrel ), 0 )
    addNC  = \(fcts, regs, pc, nc) -> (fcts, regs, pc, ((p,q):nc))
    actrel = [ (i,[[0,1]]) | i <- (agents model) ]

update model (Create prp agent n) = productUpdate model action
  where
    pre    = G (Neg (PrpF prp))
    action = ( ( [0,1], [ (0, pre   ), (1, pre   ) ],
                        [ (0, addFct), (1, addReg) ], actrel ), 0 )
    addFct = \(fcts,reg,pc,nc) -> (prp:fcts,(prp,(n, n,[ ])):reg,pc,nc)
    addReg = \(fcts,reg,pc,nc) -> (    fcts,(prp,(1,10,[n])):reg,pc,nc)
    others = delete agent (agents model)
    actrel = [ (agent,[[0],[1]]) ] ++ [ (i,[[0,1]]) | i <- others ]

update model (comA :- comB) = update (update model comA) comB

rew :: Form -> Form
rew (Top)         = Top
rew (PrpF q)      = PrpF q
rew (Equal p e)   = Equal p e
rew (Neg formula) = Neg (rew formula)
rew (Conj forms)  = Conj (map rew forms)
rew (K i formula) = K i (rew formula)
rew (G formula)   = G (rew formula)

rew (Com (AnnounceEqual p e) Top)          = Equal p e
rew (Com (AnnounceEqual p e) (PrpF q))     = Conj [Equal p e, PrpF q]
rew (Com (AnnounceEqual _ _) (Equal q f))  = Equal q f
rew (Com (AnnounceEqual p e) (Neg form))   =
  Conj [Equal p e, Neg $ (Com (AnnounceEqual p e) form)]
rew (Com (AnnounceEqual p e) (Conj forms)) =
  Conj (map (\f -> (Com (AnnounceEqual p e) f)) forms)
rew (Com (AnnounceEqual p e) (K i form))   =
  Conj [Equal p e, K i (implies (Equal p e) (Com (AnnounceEqual p e) form))]
rew (Com (AnnounceEqual p e) (G form))     =
  Conj [Equal p e, G   (implies (Equal p e) (Com (AnnounceEqual p e) form))]

rew (Com (AnnounceNotEqual p e) Top)         = Neg (Equal p e)
rew (Com (AnnounceNotEqual p e) (PrpF q))    = Conj [Neg (Equal p e), PrpF q]
rew (Com (AnnounceNotEqual p e) (Equal q f)) = Conj [Neg (Equal p e), Equal q f]
rew (Com (AnnounceNotEqual p e) (Neg f))     =
  Conj [Neg (Equal p e), Neg (Com (AnnounceNotEqual p e) f)]
rew (Com (AnnounceNotEqual p e) (Conj fs))   =
  Conj (map (\f -> (Com (AnnounceNotEqual p e) f)) fs)
rew (Com (AnnounceNotEqual p e) (K i f))     =
  Conj [Neg (Equal p e), K i (implies (Neg (Equal p e)) (Com (AnnounceNotEqual p e) f))]
rew (Com (AnnounceNotEqual p e) (G f))       =
  Conj [Neg (Equal p e), G (implies (Neg (Equal p e)) (Com (AnnounceNotEqual p e) f))]

rew (Com (Create p _ _) Top )      = G (Neg (PrpF p))
rew (Com (Create p _ _) (PrpF q) ) =
  if (p==q)
    then G (Neg (PrpF p))
    else Conj [G (Neg (PrpF p)), PrpF q]
rew (Com (Create p _ n) (Equal q (Nmbr m))) =
  if (p==q)
    then
      if (n==m)
	then G (Neg (PrpF p))                          -- R3a1
	else bot                                       -- R3a1'
    else
      Conj [G (Neg (PrpF p)), Equal q (Nmbr m)]        -- R3a2
rew (Com (Create p _ n) (Equal q (PrpE r))) =
  if (p==q)
    then
      if (r==p)
	then G (Neg (PrpF p))                          -- R3b1
	else Conj [G (Neg (PrpF p)), Equal q (Nmbr n)] -- R3b1'
    else
      if (r==p)
	then Conj [G (Neg (PrpF p)), Equal q (Nmbr n)] -- R3b2
	else Conj [G (Neg (PrpF p)), Equal q (PrpE r)] -- R3b2'
rew (Com (Create p i n) (Neg form))   =
  Conj [G (Neg (PrpF p)), Neg (Com (Create p i n) form)]
rew (Com (Create p i n) (Conj forms)) =
  Conj (map (\f -> (Com (Create p i n) f)) forms)
rew (Com (Create p i n) (K j form))   =
  if (i==j)
    then Conj [G (Neg (PrpF p)), K i (implies (G (Neg (PrpF p))) (Com (Create p i n) form))]
    else Conj [G (Neg (PrpF p)), K j form]
rew (Com (Create p i n) (G form))          = G (Com (Create p i n) form)

cmdFree :: Form -> Form
cmdFree = lfp rew

ggShowProp :: Prp -> String
ggShowProp prp = replace (replace (show prp) " 0" "") " " "_"

ggShowCnstr :: [Constraint] -> [Constraint] -> String
ggShowCnstr []  []  = ""
ggShowCnstr pcs ncs = sepBy (positives ++ negatives) " \\text{ and } "
  where
    positives = map (niceCon "   =  ") pcs
    negatives = map (niceCon " \\neq ") ncs
    niceCon b (pA,pB) = " $ " ++ (ggShowProp pA) ++ b ++ (ggShowProp pB) ++ " $ "

ggShowVal :: Valuation -> String
ggShowVal (fcts,reg,pcs,ncs) = sepBy [niceprops,nicereg,(ggShowCnstr pcs ncs)] newline
  where
    nicereg = sepBy (map niceregSingle reg) newline
    niceregSingle (p,(n,m,x)) = if (n /= m)
      then "$" ++ (show n) ++ "\\leq " ++ (ggShowProp p) ++ "\\leq " ++ (show m) ++ " $ and $ "++(ggShowProp p)++"\\not\\in \\{" ++ (sepBy (map show (sort x)) ",") ++ "\\} $"
      else " $ " ++ (ggShowProp p) ++ "=" ++ (show n) ++ " $ "
    niceprops = " $ " ++ sepBy (map ggShowProp fcts) "," ++ " $ "

ggInfo :: GuessM -> String
ggInfo m = begintab ++ "Agents: " ++ (sepBy (map show (agents m)) ", ") ++ endtab

ggTexModel :: GuessM -> String -> IO String
ggTexModel model@(Mo sts rel val cur) =
  texModel show show ggShowVal (ggInfo model) (VisModel sts rel val cur)

ggDispModel :: GuessM -> IO String
ggDispModel model@(Mo sts rel val cur) =
  dispModel show show ggShowVal (ggInfo model) (VisModel sts rel val cur)

ggTexForm :: Form -> String
ggTexForm Top          = "\\top"
ggTexForm (PrpF p)     = show p
ggTexForm (Equal p e)  = show p ++ "=" ++ (ggTexExp e)
ggTexForm (Neg f)      = " \\lnot ( " ++ ggTexForm f ++ " ) "
ggTexForm (Conj forms) = "(" ++ (sepBy (map ggTexForm forms) " \\land ") ++ ")"
ggTexForm (K i f)      = " K_{\\text{" ++ show i ++ "}} " ++ "(" ++ ggTexForm f ++ ")"
ggTexForm (G f)        = " G " ++ ggTexForm f
ggTexForm (Com c f) =   "\\langle " ++ (ggTexCom c) ++ "\\rangle " ++ ggTexForm f

ggTexExp :: Exp -> String
ggTexExp (Nmbr n) = show n
ggTexExp (PrpE p) = show p

ggTexCom :: Com -> String
ggTexCom (Create p i n)         = " " ++ show p ++ " \\stackrel{"++(show i)++"}{\\leftarrow} " ++ show n ++ " "
ggTexCom (AnnounceEqual p e)    = " ! " ++ show p ++ " = " ++ (ggTexExp e) ++ " "
ggTexCom (AnnounceNotEqual p e) = " ! " ++ show p ++ " \neq " ++ (ggTexExp e) ++ " "
ggTexCom (com1 :- com2)         = ggTexCom com1 ++ " ; " ++ ggTexCom com2

ggDispForm :: Form -> IO String
ggDispForm form = dispTexCode (" \\[ " ++ (ggTexForm form) ++ " \\] ")

