module KRIPKEVIS where
import Data.List
import System.IO
import System.Process

begintab,endtab,newline :: String
begintab  = "\\\\begin{tabular}{c}"
endtab    = "\\\\end{tabular}"
newline   = " \\\\\\\\[0pt] "

type PartitionOf a = [[a]]

data VisModel a b c = VisModel [a] [(b,PartitionOf a)] [(a,c)] a

stringModel :: Ord a => Eq a => Eq b => Show a => Show b =>
  (a -> String) -> (b -> String) -> (c -> String) -> String
    -> VisModel a b c
      -> String
stringModel showState showAgents showVal information model =
  "graph G { \n\
   \  node [shape=doublecircle,label=\""
    ++ labelof cur ++ "\"] w" ++ showState cur ++ ";\n"
    ++ concat [ ndlinefor s | s <- otherstates ]
    ++ "  rankdir=LR; \n size=\"6,5!\" \n"
    ++ concat [ "  w" ++ show x ++" -- w"++ show y
	      ++"[ label = \""++ showAgents a ++"\" ]; \n"
	    | (a,x,y) <- edges  ]
    ++ "  label = \""++information++"\"; \n "
    ++ "} \n"
  where
    (VisModel states rel val cur) = model
    labelof s = begintab ++ "\\\\textbf{" ++ showState s ++ "}" ++ newline ++ showVal (visApply val s) ++ endtab
    ndlinefor s = "  node [shape=circle,label=\""
		++ labelof s ++ "\"] w" ++ show s ++ ";\n"
    otherstates = delete cur states
    edges = nub $ concat $ concat $ concat [ [ [ if (x < y) then [(a,x,y)] else [] | x <- part, y <- part ] | part <- (visApply rel a) ] | a <- (agents model) ]
    agents (VisModel _ rel' _ _) = map fst rel'

dotModel :: Ord a => Eq a => Eq b => Show a => Show b =>
  (a -> String) -> (b -> String) -> (c -> String) -> String
    -> VisModel a b c
      -> String -> IO String
dotModel showState showAgents showVal info model filename =
  let gstring = stringModel showState showAgents showVal info model
  in do
    newFile <- openFile (filename) WriteMode
    hPutStrLn newFile gstring
    hClose newFile
    return ("Model was DOT'd to "++filename)

texModel :: Ord a => Eq a => Eq b => Show a => Show b =>
  (a -> String) -> (b -> String) -> (c -> String) -> String
    -> VisModel a b c
      -> String -> IO String
texModel showState showAgents showVal info model filename =
  do
    forget <- dotModel showState showAgents showVal info model ("tex/"++filename++".dot")
    putStrLn forget
    _ <- system ("cd tex/; dot2tex --figonly -ftikz -traw -p --autosize -w --usepdflatex "++filename++".dot > "++filename++".tex;" )
    return ("Model was TeX'd to tex/" ++ filename ++ ".tex")

dispModel :: Ord a => Eq a => Eq b => Show a => Show b =>
  (a -> String) -> (b -> String) -> (c -> String) -> String
    -> VisModel a b c
      -> IO String
dispModel showState showAgents showVal info model =
  do
    forget <- dotModel showState showAgents showVal info model "tmp/temp.dot"
    putStrLn forget
    _ <- system ("cd tmp/; dot2tex -ftikz -traw -p --autosize -w --usepdflatex temp.dot > temp.tex; pdflatex -interaction=nonstopmode temp.tex > temp.pdflatex.log; okular temp.pdf;" )
    return ("Model was TeX'd and shown.")

minimalHeader :: String
minimalHeader = "\\documentclass[12pt]{article} \n\
  \ \\usepackage[utf8]{inputenc} \n\
  \ \\usepackage[vcentering,papersize={16cm,9cm},total={15cm,8cm}]{geometry} \n\
  \ \\begin{document} \n"
minimalFooter :: String
minimalFooter = "\n \\end{document} \n"

dispTexCode :: String -> IO String
dispTexCode code = do
  newFile <- openFile ("tmp/code.tex") WriteMode
  hPutStrLn newFile (minimalHeader ++ code ++ minimalFooter)
  hClose newFile
  _ <- system ("cd tmp/; pdflatex -interaction=nonstopmode code.tex > code.pdflatex.log; okular code.pdf;" )
  return ("Code was TeX'd and shown.")

-- A helper function to apply relations:

visApply :: Show a => Eq a => [(a,b)] -> a -> b
visApply rel left =
  if (elem left ( map fst rel ))
    then snd $ head $ filter (\(a,_) -> a==left) rel
    else
      error ("Applying of a relation failed. Cannot visualize this.")