module REL where
import Data.List

-- a is a type variable, we allow all kinds of relations:
type Rel a = [(a,a)]

concatRel :: Eq a => Rel a -> Rel a -> Rel a
concatRel r s = nub [ (x,z) | (x,y) <- r, (w,z) <- s, y == w ]

lfp :: Eq a => (a -> a) -> a -> a -- least fixed point
lfp f x | x == f x  = x
	| otherwise = lfp f (f x)

dom :: Eq a => Rel a -> [a] -- "domain"
dom r = nub (foldr (\ (x,y) -> ([x,y]++)) [] r)

rtc :: Eq a => Rel a -> Rel a -- reflexive-transitive closure
rtc r = lfp (\ s -> (s `union` (concatRel r s))) i
  where xs = dom r 
	i  = [(x,x) | x <- xs ]

symc :: Eq a => Rel a -> Rel a -- symmetric closure
symc r = nub $ r ++ [ (y,x) | (x,y) <- r ]

rtsc :: Eq a => Rel a -> Rel a -- reflexive-transitive-symmetric closure
rtsc r = symc (rtc r)

-- fst and friends for triples, quadruples and quintuples:
fst3 :: (a,b,c) -> a
fst3 (a,_,_) = a

snd3 :: (a,b,c) -> b
snd3 (_,b,_) = b

trd3 :: (a,b,c) -> c
trd3 (_,_,c) = c

fst4 :: (a,b,c,d) -> a
fst4 (a,_,_,_) = a

snd4 :: (a,b,c,d) -> b
snd4 (_,b,_,_) = b

trd4 :: (a,b,c,d) -> c
trd4 (_,_,c,_) = c

fth4 :: (a,b,c,d) -> d
fth4 (_,_,_,d) = d

fst5 :: (a,b,c,d,e) -> a
fst5 (a,_,_,_,_) = a

snd5 :: (a,b,c,d,e) -> b
snd5 (_,b,_,_,_) = b

trd5 :: (a,b,c,d,e) -> c
trd5 (_,_,c,_,_) = c

fth5 :: (a,b,c,d,e) -> d
fth5 (_,_,_,d,_) = d

fft5 :: (a,b,c,d,e) -> e
fft5 (_,_,_,_,e) = e

relevant :: Ord a => Eq a => Rel a -> Rel a -- relevant part of a relation
relevant r = filter (\p -> not $ impl p) s
  where s = filter (\(x,y) -> (x < y)) r
	impl (a,b) = any (\c -> ( elem (a,c) s && elem (c,b) s )) (dom r)

apply :: Show a => Eq a => [(a,b)] -> a -> b -- apply a relation, return the first match
apply rel left = 
  if (elem left ( map fst rel ))
    then snd $ head $ filter (\(a,_) -> a==left) rel
    else
      error ("apply failed: Relation is not defined at "++(show left)++" but only at " ++ show (map fst rel) )

cartProd :: [a] -> [b] -> [(a,b)] -- Cartesian product
cartProd xs ys = [ (x,y) | x <- xs, y <- ys ]

replace :: Eq a => [a] -> [a] -> [a] -> [a] -- replace an element in a list
replace [] _ _ = []
replace s search repl =
    if take (length search) s == search
        then repl ++ (replace (drop (length search) s) search repl)
        else [head s] ++ (replace (tail s) search repl)

sepBy :: [String] -> String -> String -- separating a set of elements with a separator
sepBy []  _ = ""
sepBy [x] _ = x
sepBy (now:todo) sep = sepByStep todo sep now
sepByStep :: [String] -> String -> String -> String
sepByStep [] _ done = done
sepByStep (now:todo) sep done = sepByStep todo sep (done ++ sep ++ now)
